/******************************************************************************
* File name     : Subsequence Matching Module
* Date          : December 6, 2005
* Note          : The two functions (subMatch_best, subMatch_epsilon) find 
*               : subsequences similar to a given query sequence in a dynamic 
*               : programming sense.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "submatch.h"

/******************************************************************************
* Function name : subMatch_malloc
* Date          : December 6, 2005
* Input         : size_t      buffer_size1  ... Length of Sequence #1
* Input         : size_t      buffer_size2  ... Length of Sequence #2
* Output        : dp_match_t * != NULL      ... Address of the management table
*               :              == NULL      ... Memory allocation failure
* Note          : 
******************************************************************************/

dp_match_t* subMatch_malloc(
  size_t      buffer_size1,
  size_t      buffer_size2
){

  dp_match_t* self;                    /* Management table                   */
  struct _dp_match_opt_value *buffer;  /* Results of distance computation    */
  size_t total_buffer_size;            /* Total buffer size                  */
  struct _opt_subseq *buf_subseq;      /* Query result                       */

  /*
   * Memory allocation for the management table
   */
  self = (dp_match_t *)malloc(sizeof(dp_match_t));
  if(self == NULL){
    printf("Out of memory in subMatch_malloc\n");
    exit(1);
  }

  /*
   * Results of distance computation
   */
  total_buffer_size = buffer_size2 * 2;
  buffer = (struct _dp_match_opt_value *)malloc(
    sizeof(struct _dp_match_opt_value) * total_buffer_size);
  if(buffer == NULL){
    free(self);
    printf("Out of memory in subMatch_malloc\n");
    exit(1);
  }                       
  self->result1 = buffer;
  self->result2 = buffer + buffer_size2;

  /*
   * Final query result
   * Used only for the naive method
   */
  total_buffer_size = buffer_size1;
  buf_subseq = (struct _opt_subseq *)malloc(
    sizeof(struct _opt_subseq) * total_buffer_size);
  if(buf_subseq == NULL){
    free(self->result1);
    free(self);
    printf("Out of memory in subMatch_malloc\n");
    exit(1);
  }                       
  self->ans = buf_subseq;

  return self;
}

/******************************************************************************
* Function name : subMatch_clear
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Note          : Memory release
******************************************************************************/

void subMatch_clear(
  dp_match_t* self
){
  free(self->ans);     // Used only for the naive method

  free(self->result1);
  free(self);
}


/******************************************************************************
* Function name : subMatch_pad_disjoint
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : float       *sequence1    ... Sequence #1
* Input         : float       *sequence2    ... Sequence #2
* Input         : float       epsilon       ... Threshold
* Note          : Find the subsequence whose distance from a given sequence
*               : is the smallest. 
*               :      seq1
*               :     +-----------
*               :    s|  |
*               :    e|  |  Compute distances from upper to lower
*               :    q|  V
*               :    2|
*               : The result of the distance computation of the grid point
*               : (i,j) is stored in `res_cur[j]'. `res_pri[j]' is the result 
*               : of (i-1,j). The result of the grid point (i,j) is: 
*               : 
*               : res_cur[j] = D(sequence1[i], sequence2[j]) 
*               :                + min(res_pri[j-1], res_cur[j-1], res_pri[j])
*               : 
******************************************************************************/

void subMatch_pad_disjoint(
  dp_match_t* self,
  float       *sequence1,
  float       *sequence2,
  float       epsilon
){

  int         length1;                 /* Length of Sequence #1              */
  int         length2;                 /* Length of Sequence #2              */
  int         len2m, end_point, i;

  float       dist_cur;              /* Distance value of the current grid   */
  float       dist_opt;              /* Optimum distance of the current grid */
  long int    pos1;                    /* Cursor of Sequence #1              */
  long int    pos2;                    /* Cursor of Sequence #2              */

  struct _dp_match_opt_value *res_cur; /* Results of the current column      */
  struct _dp_match_opt_value *res_pri; /* Results of the prior (past) column */
  struct _dp_match_opt_value *swap;    /* For pointer-swapping               */
  struct _opt_subseq ans[1];           /* Candidate subsequence              */

  float      *sp1;                     /* Pointer for Sequence #1            */
  float      *sp2;                     /* Pointer for Sequence #2            */

  /*
   * Initialization
   */
  res_cur = self->result1;
  res_pri = self->result2;
  length1 = self->seq1_length;
  length2 = self->seq2_length;
  len2m = length2 - 1;
  ans->distance = INITIALIZE_VALUE;

  /*
   * Compute the results of the first column
   */
  res_pri[0].optimal_value = CALC_DISTANCE(sequence1[0], sequence2[0])
                               * DIAGONAL_WEIGHT;
  res_pri[0].column_number = 0;

  for(pos2 = 1, sp2 = sequence2 + 1; pos2 < length2; pos2++, sp2++){
    /*
     * Only use the vertical direction for the first column
     */
    res_pri[pos2].optimal_value = res_pri[pos2 - 1].optimal_value
      + CALC_DISTANCE(sequence1[0], sequence2[pos2]);
    res_pri[pos2].column_number = 0;
  }

  if (res_pri[len2m].optimal_value <= epsilon){
    ans->distance = res_pri[len2m].optimal_value;
    ans->start_point = 0;
    ans->length = 1;
  }

  /*
   * From the second column to the end column
   */
  for(pos1 = 1, sp1 = sequence1 + 1; pos1 < length1; pos1++, sp1++){

    res_cur[0].optimal_value = CALC_DISTANCE(sequence1[pos1], sequence2[0])
                                 * DIAGONAL_WEIGHT;
    res_cur[0].column_number = pos1;

    for(pos2 = 1, sp2 = sequence2 + 1; pos2 < length2; pos2++, sp2++){

      /*
       * Check the three directions, diagonal, vertical, horizontal ones, 
       * then choose the smallest among these values. 
       */
      res_cur[pos2].optimal_value = INITIALIZE_VALUE;
                                 /* Initialize the distance value            */
      dist_cur = CALC_DISTANCE(*sp1, *sp2);
                                 /* Compute the distance of the current grid */
      /*
       * Diagonal direction
       */
      res_cur[pos2].optimal_value 
	= res_pri[pos2 - 1].optimal_value + dist_cur * DIAGONAL_WEIGHT;
      res_cur[pos2].column_number = res_pri[pos2 - 1].column_number;

      /*
       * Vertical direction
       */
      dist_opt = res_cur[pos2 - 1].optimal_value + dist_cur;
      if(res_cur[pos2].optimal_value > dist_opt){
	res_cur[pos2].optimal_value = dist_opt;
	res_cur[pos2].column_number = res_cur[pos2 - 1].column_number;
      }

      /*
       * Horizontal direction
       */
      dist_opt = res_pri[pos2].optimal_value + dist_cur;
      if(res_cur[pos2].optimal_value > dist_opt){
        res_cur[pos2].optimal_value = dist_opt;
	res_cur[pos2].column_number = res_pri[pos2].column_number;
      }

    }

#if OPTIMAL
    if (ans->distance <= epsilon){
      end_point = ans->start_point + ans->length - 1;
      for(i = 0; i < length2; i++)
	if (res_cur[i].optimal_value <= ans->distance
	    && res_cur[i].column_number <= end_point)
	  break;
      if (i == length2){
#if RESULT
	printf("time:%d, start:%d, length:%d, distance:%f\n", 
	       pos1, ans->start_point, ans->length, ans->distance);
#endif
	ans->distance = INITIALIZE_VALUE;
	for(pos2 = 0; pos2 < length2; pos2++)
	  if (res_cur[pos2].column_number <= end_point)
	    res_cur[pos2].optimal_value = INITIALIZE_VALUE;
      }
    }
    if (res_cur[len2m].optimal_value <= epsilon
	&& res_cur[len2m].optimal_value < ans->distance){
      ans->distance = res_cur[len2m].optimal_value;
      ans->start_point = res_cur[len2m].column_number;
      ans->length = pos1 - res_cur[len2m].column_number + 1;
    }
#else
    if (res_cur[len2m].optimal_value <= epsilon){
#if RESULT
      printf("time:%d, start:%d, length:%d, distance:%f\n", 
	     pos1, res_cur[len2m].column_number, 
	     pos1 - res_cur[len2m].column_number + 1, 
	     res_cur[len2m].optimal_value);
#endif
      for(pos2 = 0; pos2 < length2; pos2++)
	res_cur[pos2].optimal_value = INITIALIZE_VALUE;
    }
#endif

    /*
     * Swap the results of the distance computation
     */
    swap    = res_cur;
    res_cur = res_pri;
    res_pri = swap;
  }

#if OPTIMAL
#if RESULT
  if (ans->distance <= epsilon)
    printf("time:%d start:%d, length:%d, distance:%f\n", 
	   pos1, ans->start_point, ans->length, ans->distance);
#endif
#endif
}

/******************************************************************************
* Function name : subMatch_cubic_disjoint
* Date          : December 6, 2005
* Input         : dp_match_t* self          ... Address of the management table
* Input         : float       *sequence1    ... Sequence #1
* Input         : float       *sequence2    ... Sequence #2
* Input         : float       epsilon       ... Threshold
* Note          : Find the best subsequences by the naive solution. 
******************************************************************************/

void subMatch_cubic_disjoint(
  dp_match_t* self,
  float       *sequence1,
  float       *sequence2,
  float       epsilon
){

  int         length1;                 /* Length of Sequence #1              */
  int         length2;                 /* Length of Sequence #2              */
  int         len2m, ssm;
  int         area_end = -1;           /* End point of overlapped area       */

  float       dist_cur;              /* Distance value of the current grid   */
  float       dist_opt;              /* Optimum distance of the current grid */
  long int    pos0;                    /* Cursor of subsequence matching     */
  long int    pos1;                    /* Cursor of Sequence #1              */
  long int    pos2;                    /* Cursor of Sequence #2              */

  struct _dp_match_opt_value *res_cur; /* Results of the current column      */
  struct _dp_match_opt_value *res_pri; /* Results of the prior (past) column */
  struct _dp_match_opt_value *swap;    /* For pointer-swapping               */

  float      *sp1;                     /* Pointer for Sequence #1            */
  float      *sp2;                     /* Pointer for Sequence #2            */

  /*
   * Initialization
   */
  res_cur = self->result1;
  res_pri = self->result2;
  length1 = self->seq1_length;
  length2 = self->seq2_length;
  len2m = length2 - 1;
  self->subseqs = 0;

  for (pos0 = 0 ; pos0 < length1 ; pos0++){

    /*
     * Compute the results of the first column
     */
    res_pri[0].optimal_value = CALC_DISTANCE(sequence1[pos0], sequence2[0])
                                    * DIAGONAL_WEIGHT;

    for(pos2 = 1, sp2 = sequence2 + 1; pos2 < length2; pos2++, sp2++){
      /*
       * Only use the vertical direction for the first column
       */
      res_pri[pos2].optimal_value = res_pri[pos2 - 1].optimal_value
	+ CALC_DISTANCE(sequence1[pos0], sequence2[pos2]);
    }

    if (res_pri[len2m].optimal_value <= epsilon){
      if (pos0 > area_end){
	ssm = self->subseqs;
	self->ans[ssm].distance = res_pri[len2m].optimal_value;
	self->ans[ssm].start_point = pos0;
	self->ans[ssm].length = 1;
	self->subseqs++;
	area_end = pos0;
      }
      else{
	ssm = self->subseqs - 1;
	if (self->ans[ssm].distance > res_pri[len2m].optimal_value){
	  self->ans[ssm].distance = res_pri[len2m].optimal_value;
	  self->ans[ssm].start_point = pos0;
	  self->ans[ssm].length = 1;
	  if (area_end < pos0)
	    area_end = pos0;
	}
      }
    }

    /*
     * From the second column to the end column
     */
    for(pos1 = pos0+1, sp1 = sequence1 +pos0+1; pos1 < length1; pos1++, sp1++){

      for(pos2 = 0, sp2 = sequence2; pos2 < length2; pos2++, sp2++){

	/*
	 * Check the three directions, diagonal, vertical, horizontal ones, 
	 * then choose the smallest among these values. 
	 */
	res_cur[pos2].optimal_value = INITIALIZE_VALUE;
                                 /* Initialize the distance value            */
	dist_cur = CALC_DISTANCE(*sp1, *sp2);
                                 /* Compute the distance of the current grid */

	if (pos2 > 0){
	  /*
	   * Diagonal direction
	   */
	  res_cur[pos2].optimal_value 
	    = res_pri[pos2 - 1].optimal_value + dist_cur * DIAGONAL_WEIGHT;

	  /*
	   * Vertical direction
	   */
	  dist_opt = res_cur[pos2 - 1].optimal_value + dist_cur;
	  if(res_cur[pos2].optimal_value > dist_opt)
	    res_cur[pos2].optimal_value = dist_opt;
	}

	/*
	 * Horizontal direction
	 */
	dist_opt = res_pri[pos2].optimal_value + dist_cur;
	if(res_cur[pos2].optimal_value > dist_opt)
	  res_cur[pos2].optimal_value = dist_opt;
      }

      /*
       * Swap the results of the distance computation
       */
      swap    = res_cur;
      res_cur = res_pri;
      res_pri = swap;

      if (res_pri[len2m].optimal_value <= epsilon){
	if (pos0 > area_end){
	  ssm = self->subseqs;
	  self->ans[ssm].distance = res_pri[len2m].optimal_value;
	  self->ans[ssm].start_point = pos0;
	  self->ans[ssm].length = pos1 - pos0 + 1;
	  self->subseqs++;
	  area_end = pos1;
	}
	else{
	  ssm = self->subseqs - 1;
	  if (self->ans[ssm].distance > res_pri[len2m].optimal_value){
	    self->ans[ssm].distance = res_pri[len2m].optimal_value;
	    self->ans[ssm].start_point = pos0;
	    self->ans[ssm].length = pos1 - pos0 + 1;
	  }
	  if (area_end < pos1)
	    area_end = pos1;
	}
      }
    }
  }
}
